/* Copyright (c) 2022-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#define LUA_CORE
#define setjmp_e2k_c

#ifdef __e2k__
#ifdef _KERNEL

#include <asm/cpu_regs.h>
#include <asm/process.h>

#define JMP_BUF_CNT 10
typedef struct _label_t { long long unsigned val[JMP_BUF_CNT]; } label_t;

struct jump_buf_e2k {
        e2k_pcsp_lo_t pcsp_lo;
        e2k_pcsp_hi_t pcsp_hi;
        e2k_psp_lo_t psp_lo;
        e2k_psp_hi_t psp_hi;
        e2k_usd_lo_t usd_lo;
        e2k_usd_hi_t usd_hi;
        e2k_sbr_t sbr;
        e2k_cr0_hi_t cr0_hi;
        e2k_cr1_lo_t cr1_lo;
        e2k_cr1_hi_t cr1_hi;
};

__always_inline
static void sjmp(struct jump_buf_e2k *jb)
{
	unsigned long flags;
	e2k_pcsp_lo_t pcsp_lo;
	e2k_pcsp_hi_t pcsp_hi;
	e2k_psp_lo_t psp_lo;
	e2k_psp_hi_t psp_hi;
	e2k_usd_lo_t usd_lo;
	e2k_usd_hi_t usd_hi;
	e2k_sbr_t sbr;
	e2k_cr0_hi_t cr0_hi;
	e2k_cr1_lo_t cr1_lo;
	e2k_cr1_hi_t cr1_hi;
	e2k_pshtp_t pshtp;
	e2k_pcshtp_t pcshtp;

	raw_all_irq_save(flags);
	pcsp_lo = READ_PCSP_LO_REG();
	pcsp_hi = READ_PCSP_HI_REG();
	psp_lo = READ_PSP_LO_REG();
	psp_hi = READ_PSP_HI_REG();
	usd_lo = READ_USD_LO_REG();
	usd_hi = READ_USD_HI_REG();
	sbr = READ_SBR_REG();
	cr0_hi = READ_CR0_HI_REG();
	cr1_lo = READ_CR1_LO_REG();
	cr1_hi = READ_CR1_HI_REG();
	pshtp = READ_PSHTP_REG();
	pcshtp = READ_PCSHTP_REG_SVALUE();
	raw_all_irq_restore(flags);

	AS(psp_hi).ind += GET_PSHTP_MEM_INDEX(pshtp);
	AS(pcsp_hi).ind += PCSHTP_SIGN_EXTEND(pcshtp);

	jb->pcsp_lo = pcsp_lo;
	jb->pcsp_hi = pcsp_hi;
	jb->psp_lo = psp_lo;
	jb->psp_hi = psp_hi;
	jb->usd_lo = usd_lo;
	jb->usd_hi = usd_hi;
	jb->sbr = sbr;
	jb->cr0_hi = cr0_hi;
	jb->cr1_lo = cr1_lo;
	jb->cr1_hi = cr1_hi;
}

/* Use __interrupt to make sure that its parent's values of USD that are read */
__interrupt noinline
int setjmp_e2k (label_t *jb)
{
	sjmp((struct jump_buf_e2k *) jb);

	return 0;
}


__always_inline
static void ljmp (struct jump_buf_e2k *jb)
{
	unsigned long flags;

	BUG_ON(AS(jb->pcsp_lo).base != AS(READ_PCSP_LO_REG()).base ||
	       AS(jb->psp_lo).base != AS(READ_PSP_LO_REG()).base);

	raw_all_irq_save(flags);
	E2K_FLUSHCPU;
	WRITE_PSP_REG(jb->psp_hi, jb->psp_lo);
	WRITE_PCSP_REG(jb->pcsp_hi, jb->pcsp_lo);
	WRITE_USBR_USD_REG(jb->sbr, jb->usd_hi, jb->usd_lo);
	WRITE_CR0_HI_REG(jb->cr0_hi);
	WRITE_CR1_LO_REG(jb->cr1_lo);
	WRITE_CR1_HI_REG(jb->cr1_hi);
	raw_all_irq_restore(flags);
}

/* Use __interrupt to make sure usd.lo/usd.hi are written correctly */
__interrupt noinline
void longjmp_e2k (label_t *jb)
{
	ljmp ((struct jump_buf_e2k *) jb);
        // E2K_RETURN copy from <asm/e2k_api.h>, rval=1
	do { \
        asm volatile(   "{\n" \
                        "return %%ctpr3\n" \
                        "addd %[r0], 0, %%dr0\n" \
                        "}\n" \
                        "{\n" \
                        "ct %%ctpr3\n" \
                        "}\n" \
                        :: [r0] "ir" (1) \
                        : "ctpr3"); \
        } while (0);
}

#endif // _KERNEL
#endif // __e2k__
